/* BASED ON PASCAL SOURCE: Program Flip_Build_Map ;

   Flipping the map isn't 'SO' simple as I thought.
   Walls must go clockwise. For example, in a square room with walls no. 0-3,
   wall 1 must be right to wall 0, wall 2 must be right to wall 1 and so on.
   If just trying to flip, so, for example, wall 1 is left to wall 0, then
   problems occur (The inside of the sector is...outside...and vice-versa...)
   so we should also flip the NUMBERS of the walls, so they'll still be
   sorted in the sector clockwise.
   We should also keep an array which contains for each sector the number of
   its first wall and the number of walls inside it. This is important for
   the value nextwall for each wall, which is then determined by nextsector.
   The first wall in each sector will stay the first wall for things like
   slopes and moving things.
   To make things faster we'll also make an opposite array, which contains
   for each wall the number of sector that it's part of it.
   Another array is needed to keep all the walls data.
   Also, for a DN3D effect, we should have an array of all sectors' lotags.
   That's because when we'll reach a SECTOREFFECTOR sprite (picnum=1) with
   a lotag of 0, we can't determine it's angle (by its previous angle)
   without its sector's lotag.
   If lotag is 30, then the sprite's palette lookup would determine the
   direction to rotate (clockwise or not), else it should be done by another
   sector effector with a lotag of 1 (the center one).
*/
#include <stdio.h>
#include <stdlib.h>

/* inttypes.h might be missing when using DJGPP! */
#ifdef DJGPP
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef signed char int8_t;
typedef signed short int16_t;
typedef signed int int32_t;
#else
#include <inttypes.h>
#endif

#include <string.h>
/* Can help with byte-order issues, or at least make the program a little */
/* faster on certain kinds of Little-Endian systems.                      */
/* Should be uncommented if not available.                                */
#ifdef DJGPP
  #define MACHINE_IS_LITTLE_ENDIAN
#else
  #include <sys/param.h>

  #ifdef __BYTE_ORDER
    #if __BYTE_ORDER == __LITTLE_ENDIAN
      #define MACHINE_IS_LITTLE_ENDIAN
    #endif
  #elif defined BYTE_ORDER
    #if BYTE_ORDER == LITTLE_ENDIAN
      #define MACHINE_IS_LITTLE_ENDIAN
    #endif
  #endif
#endif

#define max_sectors 32767
#define max_walls 32767

#define SECTOREFFECTOR 1
#define CYCLER 7
#define VIEWSCREEN2 499
#define VIEWSCREEN 502

typedef struct
{
  int16_t wallptr, wallnum;
} sector_walls_type;

typedef sector_walls_type sectors_walls_type[max_sectors];
typedef int16_t walls_sectors_type[max_walls];
typedef int16_t sectors_lotags_type[max_sectors];
/* BASED ON CODE FROM build.h, PART OF THE BUILD ENGINE SOURCE CODE. */
        /*32 bytes*/
typedef struct
{
        int32_t x, y;
        int16_t point2, nextwall, nextsector, cstat;
        int16_t picnum, overpicnum;
        int8_t shade;
        uint8_t pal, xrepeat, yrepeat, xpanning, ypanning;
        int16_t lotag, hitag, extra;
} wall_type;

typedef wall_type walls_type[max_walls];

typedef wall_type *pwall_type;
typedef walls_type *pwalls_type;
typedef sectors_walls_type *psectors_walls_type;
typedef walls_sectors_type *pwalls_sectors_type;
typedef sectors_lotags_type *psectors_lotags_type;
typedef int8_t *pint8_t;
typedef int16_t *pint16_t;
typedef int32_t *pint32_t;
typedef uint16_t *puint16_t;
typedef uint32_t *puint32_t;

/****************************************************************************/
/* s = signed; le = little-endian. */

void fread_s16bits_le(pint16_t p_num, FILE *f_input)
{
#ifdef MACHINE_IS_LITTLE_ENDIAN
  fread(p_num, 2, 1, f_input);
#else
  uint8_t a[2];
  fread(a, 2, 1, f_input);
  *p_num = (int16_t)((a[1]<<8)|a[0]);
#endif
}

/****************************************************************************/

void fread_s32bits_le(pint32_t p_num, FILE *f_input)
{
#ifdef MACHINE_IS_LITTLE_ENDIAN
  fread(p_num, 4, 1, f_input);
#else
  uint8_t a[4];
  fread(a, 4, 1, f_input);
  *p_num = (int32_t)((a[3]<<24)|(a[2]<<16)|(a[1]<<8)|a[0]);
#endif
}

/****************************************************************************/

void fwrite_s16bits_le(pint16_t p_num, FILE *f_output)
{
#ifdef MACHINE_IS_LITTLE_ENDIAN
  fwrite(p_num, 2, 1, f_output);
#else
  uint8_t a[2];
  a[0] = ((uint16_t)(*p_num)) & 0xFF;
  a[1] = ((uint16_t)(*p_num)) >> 8;
  fwrite(a, 2, 1, f_output);
#endif
}

/****************************************************************************/

void fwrite_s32bits_le(pint32_t p_num, FILE *f_output)
{
#ifdef MACHINE_IS_LITTLE_ENDIAN
  fwrite(p_num, 4, 1, f_output);
#else
  uint8_t a[4];
  a[0] = ((uint32_t)(*p_num)) & 0xFF;
  a[1] = (((uint32_t)(*p_num)) >> 8) & 0xFF;
  a[2] = (((uint32_t)(*p_num)) >> 16) & 0xFF;
  a[3] = ((uint32_t)(*p_num)) >> 24;
  fwrite(a, 4, 1, f_output);
#endif
}

/****************************************************************************/

void write_common_wall_data(FILE *f_output, pwall_type wall)
{
  fwrite_s16bits_le(&wall->nextsector, f_output);
  wall->cstat ^= 8;
  fwrite_s16bits_le(&wall->cstat, f_output);
  fwrite_s16bits_le(&wall->picnum, f_output);
  fwrite_s16bits_le(&wall->overpicnum, f_output);
  fwrite(&wall->shade, 1, 1, f_output);
  fwrite(&wall->pal, 1, 1, f_output);
  fwrite(&wall->xrepeat, 1, 1, f_output);
  fwrite(&wall->yrepeat, 1, 1, f_output);
  fwrite(&wall->xpanning, 1, 1, f_output);
  fwrite(&wall->ypanning, 1, 1, f_output);
  fwrite_s16bits_le(&wall->lotag, f_output);
  fwrite_s16bits_le(&wall->hitag, f_output);
  fwrite_s16bits_le(&wall->extra, f_output);
}

/****************************************************************************/

/* The procedure gets a pointer pwalls to walls and pk, a pointer
   to the first wall in a circle of walls inside a sector.
   The procedure returns via pnum the number of walls in the circle.
   Pointers to the given data are used to decrease memory usage.            */

void get_num_walls_in_circle(pwalls_type walls, pint16_t k, pint16_t num)
{
  /* tempwallnum is used to get the starting positions of other walls. */
  int16_t tempwallnum = *k;
  *num = 1;
  do
  {
    (*num)++;
    tempwallnum = ((*walls)[tempwallnum]).point2;
  }
  while (((*walls)[tempwallnum]).point2 != *k);
}

/****************************************************************************/

/* The procedure gets a pointer pwalls to walls and pk, a pointer
   to a wall in a circle of walls inside a sector.
   In addition, pointers psectors_walls and pwalls_sectors are used to match
   between a wall and a sector, each time the opposite way.
   The procedure returns via pfirst the number of the first wall in the
   circle, and via pnum the number of walls in the circle.
   Pointers to the given data are used to decrease memory usage.            */

void get_num_walls_of_wall_in_circle(pwalls_type walls, psectors_walls_type sectors_walls,
                                     pwalls_sectors_type walls_sectors, pint16_t k, pint16_t first, pint16_t num)
{
  int8_t not_found;
  int16_t current = ((*sectors_walls)[(*walls_sectors)[*k]]).wallptr;
  do
  {
    *first = current;
    not_found = (current != *k);
    while ((not_found) && (*first != ((*walls)[current]).point2))
    {
      current = ((*walls)[current]).point2;
      not_found = (current != *k);
    }
    current++;
  }
  while (not_found);
  get_num_walls_in_circle(walls, first, num);
}

int main(int argc, char **argv)
{
  FILE *f_input, *f_output;
  uint8_t params_bits;
  uint8_t a_byte;
  int16_t a_word, tempword1, tempword2, tempword3, n_things,
        i, j, k, num, nextfirstincircle, next_num;
  int32_t a_long, templong1, templong2;
  psectors_walls_type sectors_walls;
  pwalls_sectors_type walls_sectors;
  psectors_lotags_type sectors_lotags = NULL;
  pwalls_type walls;
  if (argc < 3)
  {
    printf("\nTHE BUILD/DUKE NUKEM 3D MAP FLIPPER V0.92 BY NY00123\n");
    printf("Using it is simple. Just type in the command line:\n");
    printf("flip_map mymap.map flipped.map [d or v] (or ./flip_map in Linux/Unix if needed)\n");
    printf("mymap.map represents a certain Build/DN3D map (level) file,\n");
    printf("and flipped.map represents the file to be created for the flipped version.\n");
    printf("The third parameter is optional. If it's d, when converting a DN3D map,\n");
    printf("it would take care of DN3D-specific effects such as rotating doors.\n");
    printf("If it's v, flipping will be like d, but viewscreens won't be flipped,\n");
    printf("so when watching a viewscreen you should not get a flipped photography.\n");
    printf("Full paths are allowed. Examples when using a Windows/DOS command line:\n");
    printf("flip_map c:\\mymaps\\buildmap.map flipped.map\n");
    printf("flip_map mydn3dmap.map d:\\flipmaps\\dn3dmapflipped.map d\n");
    printf("flip_map c:\\dukemaps\\mymap.map d:\\flipmaps\\flipped.map v\n");
    printf("Examples with a Linux/Unix command line (depending on where is flip_map):\n");
    printf("./flip_map mydn3dmap.map ~/flipmaps/dn3dmapflipped.map d\n");
    printf("flip_map /home/myfriend/dukemaps/mymap.map ~/flipmaps/flipped.map v\n\n");
    printf("This utility, including the source code, is free of charge and comes\n");
    printf("WITHOUT ANY KIND OF WARRANTY! USE AT YOUR OWN RISK!\n");
    printf("The Build MAP format is used, so flip_map is licensed under the BUILDLIC,\n");
    printf("however it contains a few DN3D-related definitions, so it's also GPLed.\n");

    return 1;
  }
  f_input = fopen(argv[1],"rb");
  if (f_input==NULL)
  {
    printf("ERROR WHILE OPENING FILE FOR INPUT.\n");
    return EXIT_FAILURE;
/*  return 2; */
  }
  f_output = fopen(argv[2],"wb");
  if (f_output==NULL)
  {
    printf("ERROR WHILE TRYING TO CREATE FILE FOR OUTPUT.\n");
    return EXIT_FAILURE;
/*  return 3; */
  }

  params_bits = (argc >= 4);
  if (params_bits)
    params_bits = ((!strcmp(argv[3], "d")) | (!strcmp(argv[3], "v") << 1));

  fread_s32bits_le(&a_long, f_input);   /* Map version. For big levels made */
  fwrite_s32bits_le(&a_long, f_output); /* with Mapster32 it's saved.       */

  fread_s32bits_le(&a_long, f_input);   /* Player Start Pos X */
  fwrite_s32bits_le(&a_long, f_output);
  fread_s32bits_le(&a_long, f_input);   /* Player Start Pos Y */
  a_long = -a_long;
  fwrite_s32bits_le(&a_long, f_output);
  fread_s32bits_le(&a_long, f_input);   /* Player Start Pos Z */
  fwrite_s32bits_le(&a_long, f_output);

  fread_s16bits_le(&a_word, f_input);   /* Player Start Pos Angle */
  if (a_word) /* Not a zero */
     a_word = 2048-a_word;
  fwrite_s16bits_le(&a_word, f_output);

  fread_s16bits_le(&a_word, f_input);   /* Player Start Pos Sector Number */
  fwrite_s16bits_le(&a_word, f_output);

  fread_s16bits_le(&n_things, f_input); /* Number of sectors */
  fwrite_s16bits_le(&n_things, f_output);

  sectors_walls = malloc(n_things*sizeof(sector_walls_type));
  if (sectors_walls == NULL)
  {
    printf("NOT ENOUGH MEMORY AVAILABLE.\n");
    return EXIT_FAILURE;
  }
  walls_sectors = malloc(sizeof(walls_sectors_type));
  if (walls_sectors == NULL)
  {
    printf("NOT ENOUGH MEMORY AVAILABLE.\n");
    free(sectors_walls);
    return EXIT_FAILURE;
  }
  if (params_bits)
  {
    sectors_lotags = malloc(n_things*sizeof(int16_t));
    if (sectors_lotags == NULL)
    {
      printf("NOT ENOUGH MEMORY AVAILABLE.\n");
      free(sectors_walls);
      free(walls_sectors);
      return EXIT_FAILURE;
    }
  }
  n_things--; /* Now minus one... */

  for(i=0;i<=n_things;i++) /* MINUS ONE... */
  {
    fread_s16bits_le(&(*sectors_walls)[i].wallptr, f_input); /* First wall */
    fwrite_s16bits_le(&(*sectors_walls)[i].wallptr, f_output);
    fread_s16bits_le(&(*sectors_walls)[i].wallnum, f_input); /* The number of walls. */
    fwrite_s16bits_le(&(*sectors_walls)[i].wallnum, f_output);
    /* a_word = (*sectors_walls)[i].wallnum - 1 + (sectors_walls)[i].wallptr ; */
    a_word = (*sectors_walls)[i].wallnum;
    a_word--;
    a_word += (*sectors_walls)[i].wallptr;
    /* WARNING:
       If there's a wall numbered 32767, though currently the Build engine
       shouldn't allow that, this might exceed the array's range and
       crash, or even worse: Write to random memory locations and continue
       in an infinite loop.                                                */
    for(j=(*sectors_walls)[i].wallptr;j<=a_word;j++)
      (*walls_sectors)[j] = i;
    fread_s32bits_le(&a_long, f_input); /* Ceiling height. */
    fwrite_s32bits_le(&a_long, f_output);
    fread_s32bits_le(&a_long, f_input); /* Floor height. */
    fwrite_s32bits_le(&a_long, f_output);

    fread_s16bits_le(&a_word, f_input); /* Ceiling stats */

    /* CAREFUL WITH STATS: If stats have bit 2 on (+4), which
       swaps between X&Y, then bit 5 (y-flip) shouldn't be flipped,
       but bit 4 (x-flip) should be flipped.
       We can use an IF statement or a 'trick' of calculations,
       however the trick might take more time than the IF statement
       so we use here the IF statement.
       ...It looks like it also happens if bit 6 is on (+64), which
       means that the texture is aligned to the first wall of the
       sector.
       However if BOTH bits 2 & 6 are on, it should NOT happen.
       Looks like X&Y swap is flipped in someway after flipping map
       and so is the texture aligning.
       Furthermore, it seems like sometimes a certain kind of an unwanted
       panning (where the texture is simply 'moved') occurs.
       Right now I don't have a solution for this, yet.
       One bit off and one bit on (of bits 2 and 6).                      */
    if (!(a_word & 4) ^ !(a_word & 64))
      a_word ^= 16;
    else /* Two bits off or two bits on. */
      a_word ^= 32;
    fwrite_s16bits_le(&a_word, f_output);

    fread_s16bits_le(&a_word, f_input); /* Floor stats */

    /* The same here with floor stats. */

    /* One bit off and one bit on (of bits 2 and 6). */
    if (!(a_word & 4) ^ !(a_word & 64))
      a_word ^= 16;
    else /* Two bits off or two bits on. */
      a_word ^= 32;
    fwrite_s16bits_le(&a_word, f_output);

    fread_s32bits_le(&a_long, f_input); /* Ceiling picture/tile and slope. */
    fwrite_s32bits_le(&a_long, f_output);
    fread_s32bits_le(&a_long, f_input);   /* Ceiling shade, palette,      */
    fwrite_s32bits_le(&a_long, f_output); /* palette lookup and pannings. */
    fread_s32bits_le(&a_long, f_input); /* Floor picture/tile and slope. */
    fwrite_s32bits_le(&a_long, f_output);
    fread_s32bits_le(&a_long, f_input);   /* Floor shade, palette,        */
    fwrite_s32bits_le(&a_long, f_output); /* palette lookup and pannings. */
    if (params_bits)
    {
      fread_s16bits_le(&a_word, f_input); /* Visibility and filler. */
      fwrite_s16bits_le(&a_word, f_output);
      fread_s16bits_le(&(*sectors_lotags)[i], f_input); /* Lotag. */
      fwrite_s16bits_le(&(*sectors_lotags)[i], f_output);
    }
    else
    {
      fread_s32bits_le(&a_long, f_input); /* Visibility, filler and Lotag. */
      fwrite_s32bits_le(&a_long, f_output);
    }
    fread_s32bits_le(&a_long, f_input); /* Hitag and extra. */
    fwrite_s32bits_le(&a_long, f_output);
  }

  fread_s16bits_le(&j, f_input); /* Number of walls */
  fwrite_s16bits_le(&j, f_output);
  /* First we can decrease the following... */
  walls_sectors = realloc(walls_sectors, j*sizeof(int16_t));
  walls = malloc(j*sizeof(wall_type));
  if (walls == NULL)
  {
    printf("NOT ENOUGH MEMORY AVAILABLE.\n");
    free(sectors_walls);
    free(walls_sectors);
    if (params_bits)
      free(sectors_lotags);
    return EXIT_FAILURE;
  }

  j--; /* Now number of walls MINUS ONE... */

  for(i=0;i<=j;i++) /* MINUS ONE... */
  {
    fread_s32bits_le(&((*walls)[i]).x, f_input);
    fread_s32bits_le(&((*walls)[i]).y, f_input);
    fread_s16bits_le(&((*walls)[i]).point2, f_input);
    fread_s16bits_le(&((*walls)[i]).nextwall, f_input);
    fread_s16bits_le(&((*walls)[i]).nextsector, f_input);
    fread_s16bits_le(&((*walls)[i]).cstat, f_input);
    fread_s16bits_le(&((*walls)[i]).picnum, f_input);
    fread_s16bits_le(&((*walls)[i]).overpicnum, f_input);
    fread(&((*walls)[i]).shade, 1, 1, f_input);
    fread(&((*walls)[i]).pal, 1, 1, f_input);
    fread(&((*walls)[i]).xrepeat, 1, 1, f_input);
    fread(&((*walls)[i]).yrepeat, 1, 1, f_input);
    fread(&((*walls)[i]).xpanning, 1, 1, f_input);
    fread(&((*walls)[i]).ypanning, 1, 1, f_input);
    fread_s16bits_le(&((*walls)[i]).lotag, f_input);
    fread_s16bits_le(&((*walls)[i]).hitag, f_input);
    fread_s16bits_le(&((*walls)[i]).extra, f_input);
  }
  /* Now with all of the walls data ready...
     N_Things = Number of LAST sector (or number of sectors MINUS ONE...) */

  j = -1; /* This will point on the walls during the next loop. */

  for(i=0;i<=n_things;i++)
    do
    {
      /* CAREFUL HERE: THE X & Y AREN'T THE SAME AS PREVIOUSLY.
         ...Right, you think Y should be -Y because of flipping.
         Correct... BUT NOT THE ORIGINAL Y!
         I mean, after flipping the Y values, the walls will go
         anticlockwise, which is NOT GOOD. They should still go
         clockwise. So if a wall starts at (x1,y1), its point2
         wall starts at (x2,y2), and the walls are sorted
         anticlockwise, then, after making the walls be again
         clockwise, the first wall will start at (x2,y2)!

         Also, if inside a room there is, for example, one
         pillar or more, then the sides of the pillar should actually
         be sorted ANTIclockwise. But basically this is the same thing.

         With the flipping of each sector:
         a. The sides should be sorted clockwise or anticlockwise,
            depending on the case, BUT the method of doing it is the same:
            flipping the "clockwise-form".
         b. We should work on each circle of walls SEPARATELY
            (The room itself, a pillar, a hole, etc).
         c. The first wall in each circle will stay with the same number,
            so aligned and sloped flats can be converted probably.         */

      j++ ; /* Should be the first wall of a circle in a sector now. */
      k = j;
      tempword1 = k; /* tempword1 would increase for num > 2. */
      /* CAN'T I PASS POINTERS TO BIG VARIABLES LIKE walls? */
      get_num_walls_in_circle(walls, &k, &num);
      if (num <= 1)
      {
        fwrite_s32bits_le(&(*walls)[j].x, f_output); /* X */
        a_long = -(*walls)[j].y;                /* Y */
        fwrite_s32bits_le(&a_long, f_output);
        fwrite_s16bits_le(&k, f_output);        /* Point2 */
      }
      else
      {
        a_word = k+1;
        fwrite_s32bits_le(&(*walls)[a_word].x, f_output); /* X */
        a_long = -(*walls)[a_word].y;                /* Y */
        fwrite_s32bits_le(&a_long, f_output);
        fwrite_s16bits_le(&a_word, f_output);        /* Point2 */
      }
      if ((*walls)[j].nextwall == -1) /* NextWall */
        fwrite_s16bits_le(&(*walls)[j].nextwall, f_output);
      else
      {
        get_num_walls_of_wall_in_circle(walls, sectors_walls, walls_sectors,
                                        &(*walls)[j].nextwall, &nextfirstincircle, &next_num);
        if ((*walls)[j].nextwall == nextfirstincircle)
          fwrite_s16bits_le(&nextfirstincircle, f_output);
        else
        {
          a_word = 2*nextfirstincircle + next_num - (*walls)[j].nextwall;
          fwrite_s16bits_le(&a_word, f_output);
        }
      }
      write_common_wall_data(f_output, &(*walls)[j]);

      /* Now writing other walls in REVERSED order... except the last wall,
         with the "special case" of it's Point2 being WallPtr.              */

      tempword2 = k+2;

      for(j=k+num-1;j>=tempword2;j--)
      {
        /* STILL CAREFUL WITH X AND Y!

           The current wall is (*walls)[j], and the next wall (point2), which
           is actually the previous one, becomes the latest (*walls)[j] value. */

        tempword3 = tempword1; /* Previous wall becomes the next wall. */
        tempword1 = j;
        fwrite_s32bits_le(&(*walls)[tempword3].x, f_output); /* X */
        a_long = -(*walls)[tempword3].y ;               /* Y */
        fwrite_s32bits_le(&a_long, f_output);

        /* Here the number of point2 is NOT j+1. j is the "old" number of
           the current wall (anticlockwise). 2*k+num-j is the new number of
           the current wall in the CURRENT CIRCLE
           (or first wall + (last wall + 1) - old number)
           (last wall = first wall + number of walls - 1).
           For point2 we should just add 1 to the result.                   */

        /* Point2 */
        a_word = 2*k+num-j+1;
        fwrite_s16bits_le(&a_word, f_output);
        /* NextWall */
        if ((*walls)[j].nextwall == -1)
          fwrite_s16bits_le(&(*walls)[j].nextwall, f_output);
        else
        {
          get_num_walls_of_wall_in_circle(walls, sectors_walls, walls_sectors,
                                          &(*walls)[j].nextwall, &nextfirstincircle, &next_num);
          if ((*walls)[j].nextwall == nextfirstincircle)
            fwrite_s16bits_le(&nextfirstincircle, f_output);
          else
          {
            a_word = 2*nextfirstincircle+next_num-(*walls)[j].nextwall;
            fwrite_s16bits_le(&a_word, f_output);
          }
        }
        write_common_wall_data(f_output, &(*walls)[j]);
      }

      /* If there are more than 1 wall, then we're going now
         to the last wall (which isn't the first one).       */

      if (num > 1)
      {
        /* STILL CAREFUL WITH X AND Y! */
        tempword3 = tempword1;
        tempword1 = k+1;
        fwrite_s32bits_le(&(*walls)[tempword3].x, f_output);
        a_long = -(*walls)[tempword3].y;
        fwrite_s32bits_le(&a_long, f_output);
        fwrite_s16bits_le(&k, f_output); /* Point2 */
        /* NextWall */
        if ((*walls)[tempword1].nextwall == -1)
        {
          fwrite_s16bits_le(&(*walls)[tempword1].nextwall, f_output);
        }
        else
        {
          get_num_walls_of_wall_in_circle(walls, sectors_walls, walls_sectors,
                                          &(*walls)[tempword1].nextwall, &nextfirstincircle, &next_num);
          if ((*walls)[tempword1].nextwall == nextfirstincircle)
            fwrite_s16bits_le(&nextfirstincircle, f_output);
          else
          {
            a_word = 2*nextfirstincircle+next_num-(*walls)[tempword1].nextwall;
            fwrite_s16bits_le(&a_word, f_output);
          }
        }
        write_common_wall_data(f_output, &(*walls)[tempword1]);
      }
      j = k+num-1;

      /* Is it the last circle of walls in sector?
         We compare k+num-1, the number of the last wall in the current
         circle, to the number of the last wall in the whole sector.    */
    }
    while (j != (*sectors_walls)[(*walls_sectors)[j]].wallptr-1+(*sectors_walls)[(*walls_sectors)[j]].wallnum);

  fread_s16bits_le(&n_things, f_input); /* Number of sprites */
  fwrite_s16bits_le(&n_things, f_output);

  for(i=1;i<=n_things;i++)
  {
    fread_s32bits_le(&a_long, f_input);   /* x */
    fwrite_s32bits_le(&a_long, f_output);
    fread_s32bits_le(&a_long, f_input);   /* y */
    a_long = -a_long;
    fwrite_s32bits_le(&a_long, f_output);
    fread_s32bits_le(&a_long, f_input);   /* z */
    fwrite_s32bits_le(&a_long, f_output);
    fread_s16bits_le(&a_word, f_input);   /* Cstat */
    /* If we've got a viewscreen, maybe the user doesn't want it to be
       flipped, so let's wait.                                         */
    fread_s16bits_le(&j, f_input);   /* Picnum */
    /* Should we flip or not? Maybe we've got a viewscreen and the user
       doesn't want viewscreens to be flipped.                          */
    if ((params_bits != 2) || ((j != VIEWSCREEN) && (j != VIEWSCREEN2)))
      a_word ^= 4;
    /* ok, let's write. */
    fwrite_s16bits_le(&a_word, f_output);
    fwrite_s16bits_le(&j, f_output);

    fread(&a_byte, 1, 1, f_input);   /* Shade */
    fwrite(&a_byte, 1, 1, f_output);
    /* Stopping writting for awhile...
       next to write is palette lookup, which might affect angle in Duke3D. */

    fread(&a_byte, 1, 1, f_input); /* Palette lookup. */
    fread_s16bits_le(&tempword1, f_input); /* Clipdist and filler. */
    fread_s32bits_le(&a_long, f_input); /* Repeats and offsets. */
    fread_s16bits_le(&num, f_input); /* Sector number. */
    fread_s16bits_le(&tempword2, f_input); /* Status number */

    fread_s16bits_le(&a_word, f_input); /* Angle */
    fread_s32bits_le(&templong1, f_input); /* Owner and X-Velocity. */
    fread_s32bits_le(&templong2, f_input); /* Y and Z Velocity components. */
    fread_s16bits_le(&k, f_input); /* Lotag */

    if (params_bits)
    {
    /* Now dealing with palette lookup and angle...
       can change if the sprite's tilenum is SECTOREFFECTOR (PicNum == 1) or CYCLER (== 7).
       Only after all of that, the final writting is done.                                  */

      if (j == SECTOREFFECTOR)
      {
        if (k) /* Different than zero? */
        {
          if ((k != 13) && (k != 18) && (k != 21) && (k != 31) && (k != 32))
          {
            if ((k == 1) || (k == 11))
            {
              if (a_word<1024)
                a_word = 1536;
              else
                a_word = 512;
            }
            else
              if (a_word) /* Different than zero? */
                a_word = 2048-a_word;
          }
        }
        else
        {
          if ((*sectors_lotags)[num] == 30)
          {
            if (a_byte) /* Is a_byte different than zero? */
              a_byte = 0;
            else
              a_byte = 1;
          }
        }
      }
      else
        /* If we encounter a CYCLER, we shouldn't change the angle in that case.
           If not, then we change the angle only if it's different than zero. */
        if ((j != CYCLER) && (a_word))
  /*    if (a_word) -- Different than zero? */
          a_word = 2048-a_word;
    }
    else
      if (a_word) /* Different than zero? */
        a_word = 2048-a_word;
    fwrite(&a_byte, 1, 1, f_output); /* Palette lookup. */
    fwrite_s16bits_le(&tempword1, f_output); /* Clipdist and filler. */
    fwrite_s32bits_le(&a_long, f_output); /* Repeats and offsets. */
    fwrite_s16bits_le(&num, f_output); /* Sector number. */
    fwrite_s16bits_le(&tempword2, f_output); /* Status number. */
    fwrite_s16bits_le(&a_word, f_output); /* Angle */

    fwrite_s32bits_le(&templong1, f_output); /* Owner and X-Velocity. */
    fwrite_s32bits_le(&templong2, f_output); /* Y and Z Velocity components. */
    fwrite_s16bits_le(&k, f_output); /* Lotag */

    /* ok, finished using the extra temporary variables. */

    fread_s32bits_le(&a_long, f_input);   /* Hitag and extra. */
    fwrite_s32bits_le(&a_long, f_output);
  }
  fclose(f_input);
  fclose(f_output);
  free(walls);
  free(sectors_walls);
  free(walls_sectors);
  if (sectors_lotags != NULL)
    free(sectors_lotags);
  return 0;
}
